
#ifndef _FLUID_RVOICE_DSP_TABLES_H
#define _FLUID_RVOICE_DSP_TABLES_H

#define FLUID_INTERP_MAX         256
#define SINC_INTERP_ORDER 7  /* 7th order constant */
#define CUBIC_INTERP_ORDER 4 /* 4th order constant */
#define LINEAR_INTERP_ORDER 2

#endif
